package cz.mendelu.dnaAnalyser.analyse.cpg;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;



@Slf4j
@Service
public class CpgAnalyser {

    private double MIN_GC_PERCENTAGE = 0.5;
    private double MIN_OBSERVED_TO_EXPECTED_CPG = 0.6;
    private char FIRST_NUCLEOTIDE = 'C';
    private char SECOND_NUCLEOTIDE = 'G';
    private int MIN_WINDOW_SIZE = 200;
    private double MIN_ISLAND_MERGE_GAP = 100;

    public static boolean find_island_1bp_shifts(CpgWindow window) {
        while (true) {
            if (window.is_island()) {
                return true;
            }
            if (!window.expand_right()) {
                return false;
            }
            window.shrink_left();
        }
    }

    public boolean two_fold_shrinking(CpgWindow window) {
        while (!window.is_island()) {
            window.shrink_right();
            window.shrink_left();
            window.window_size -= 2;
            if (window.window_size < MIN_WINDOW_SIZE) {
                return false;
            }
        }
        return true;
    }

    public List<CpgWindow> merge_islands(List<CpgWindow> found_islands) {
        List<CpgWindow> merged_islands = new ArrayList<>();
        merged_islands.add(found_islands.get(0));
        if (found_islands.size() >= 2) {
            for (int i = 1; i < found_islands.size(); i++) {
                CpgWindow found_island = found_islands.get(i);
                if (found_island.window_begin - merged_islands.get(merged_islands.size() - 1).window_end < MIN_ISLAND_MERGE_GAP) {
                    merged_islands.get(merged_islands.size() - 1).join_window(found_island);
                } else {
                    merged_islands.add(found_island);
                }
            }
        }
        return merged_islands;
    }

    public void extend_island_window_shifts(CpgWindow window) {
        while (true) {
            boolean extended = window.expand_right_island();
            if (!extended) {
                break;
            }
        }
        window.update_gc_count();
    }

    public List<CpgWindow> find_islands(String record) {
        List<CpgWindow> found_islands = new ArrayList<>();
        int record_position = 0;
        while (true) {
            CpgWindow window = new CpgWindow(record, record_position, MIN_WINDOW_SIZE, MIN_WINDOW_SIZE, MIN_GC_PERCENTAGE, MIN_OBSERVED_TO_EXPECTED_CPG, FIRST_NUCLEOTIDE, SECOND_NUCLEOTIDE);

            boolean record_continues = find_island_1bp_shifts(window);
            if (!record_continues) {
                break;
            }
            extend_island_window_shifts(window);
            window.rollback_until_island();
            if (!two_fold_shrinking(window)) {
                record_position = window.window_begin + 1;
                continue;
            }
            found_islands.add(window);
            record_position = window.window_end;
        }

        if (found_islands.isEmpty()) {
            return new ArrayList<>();
        }
        return merge_islands(found_islands);
    }

    private void initialize(int minWindowSize, double minGcPercentage, double minObservedToExpectedCpG, int minIslandMergeGap, char firstNucleotide, char secondNucleotide) {
        this.MIN_WINDOW_SIZE = minWindowSize;
        this.MIN_GC_PERCENTAGE = minGcPercentage;
        this.MIN_OBSERVED_TO_EXPECTED_CPG = minObservedToExpectedCpG;
        this.MIN_ISLAND_MERGE_GAP = minIslandMergeGap;
        this.FIRST_NUCLEOTIDE = firstNucleotide;
        this.SECOND_NUCLEOTIDE = secondNucleotide;

    }

    public ArrayList<CpgAnalyseResult> getResults(String sequence, int minWindowSize, double minGcPercentage, double minObservedToExpectedCpG, int minIslandMergeGap, char firstNucleotide, char secondNucleotide) {
        initialize(minWindowSize, minGcPercentage, minObservedToExpectedCpG, minIslandMergeGap, firstNucleotide, secondNucleotide);
        List<CpgWindow> islands = find_islands(sequence);
        ArrayList<CpgAnalyseResult> results = new ArrayList<>();
        for(CpgWindow island : islands) {
            results.add(new CpgAnalyseResult(null, island.window_begin, island.window_end, island.record.substring(island.window_begin, island.window_begin + 20) + "...", island.getGcPerc(), island.getObsExp(), island.window_size));
        }
        return results;
    }
}
