package cz.mendelu.dnaAnalyser.analyse.rloopr;

import cz.mendelu.dnaAnalyser.jwt.JwtAuthenticationUtils;
import cz.mendelu.dnaAnalyser.jwt.JwtTokenService;
import cz.mendelu.dnaAnalyser.sequence.SequenceService;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceData;
import cz.mendelu.dnaAnalyser.sequence.data.SequenceDataService;
import cz.mendelu.dnaAnalyser.sequence.Sequence;
import cz.mendelu.dnaAnalyser.exporter.ExporterService;
import cz.mendelu.dnaAnalyser.utils.analyse.service.heatmap.Heatmap;
import cz.mendelu.dnaAnalyser.utils.controller.DeleteMapping;
import cz.mendelu.dnaAnalyser.utils.controller.GetMapping;
import cz.mendelu.dnaAnalyser.utils.controller.*;
import cz.mendelu.dnaAnalyser.utils.controller.PostMapping;
import cz.mendelu.dnaAnalyser.utils.controller.PutMapping;
import cz.mendelu.dnaAnalyser.utils.swagger.ApiPaginationParams;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import lombok.Data;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.http.ContentDisposition;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.security.core.Authentication;
import org.springframework.web.bind.annotation.*;
import springfox.documentation.annotations.ApiIgnore;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Analysis;
import cz.mendelu.dnaAnalyser.utils.analyse.data.ExporterEnums.Output;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@RestController
@Api(tags = "api/analyse/rloopr")
@RequestMapping("api/analyse/rloopr")
public class RlooprController {

    @Data
    static class RlooprRequest {
        private Short[] rizModel;
        private UUID sequence;
        private Set<String> tags;
    }

    @Data
    static class HeatMapRequest {
        private Integer sequenceLength;
        private UUID sequenceId;
    }

    @Autowired
    private ExporterService exporterService;

    @Autowired
    private SequenceService sequenceService;

    @Autowired
    private SequenceDataService sequenceDataService;

    @Autowired
    private JwtTokenService jwtTokenService;

    @Autowired
    private RlooprService rlooprService;

    @PostMapping()
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Analyse sequence for R-loops")
    public Response<Rloop> findRIZ(
            @ApiIgnore Authentication authentication,
            @RequestBody @Valid RlooprRequest request) {

        Rloop rloop = rlooprService.executeRloopAnalysis(
                Rloop.builder()
                        .sequence(sequenceService.findOne(request.getSequence()))
                        .owner(JwtAuthenticationUtils.toUser(authentication))
                        .tags(request.getTags())
                        .model(request.getRizModel())
                        .build()
        );

        return Response.response(rloop);
    }

    @GetMapping
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get page with Rloop analysis.")
    public ResponsePage<Rloop> getPageWithRloopAnalysis(
            PaginationParam paginationParam,
            SortParam sortParam,
            TagFilterParam tagFilterParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<Rloop> rlooprAnalysePage = rlooprService.findAll(pageable, tagFilterParam.getTags());
        ResponsePage<Rloop> responsePage = ResponsePage.responsePage(rlooprAnalysePage);
        rlooprAnalysePage
                .forEach(rloop -> responsePage.setDownloadToken(jwtTokenService.generateDownloadToken()));
        return responsePage;
    }


    @GetMapping("/{id}/rloops")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one Rloopr analysis by ID.")
    public ResponsePage<RlooprResult> getRlooprList(
            @PathVariable UUID id,
            Integer sequenceStart,
            Integer sequenceLength,
            PaginationParam paginationParam,
            SortParam sortParam)
    {

        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<RlooprResult> rlooprPage = rlooprService.returnRloopAnalysis(id, sequenceStart, sequenceLength, pageable);
        return ResponsePage.responsePage(rlooprPage);
    }

    @GetMapping("/{id}/analysis")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get one R-loop-R analysis by ID.")
    public Response<Rloop> getRloopAnalysis(
            @PathVariable UUID id) {
        Rloop rloop = rlooprService.findOne(id);
        Response<Rloop> response = Response.response(rloop);
        response.setDownloadToken(jwtTokenService.generateDownloadToken());

        return response;
    }

    @GetMapping("/{id}/heatmap")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get R-loop-R heatmap by Id.")
    public Heatmap getRloopRHeatmap(
            @PathVariable UUID id,
            Integer segments,
            Integer from,
            Integer to)
    {
        Heatmap heatmap = rlooprService.getHeatmap(id, segments, from, to);
        return heatmap;
    }

    @GetMapping("/tag")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiPaginationParams
    @ApiOperation("Get all tags defined for sequences")
    public ResponsePage<String> getPageWithTags(
            PaginationParam paginationParam,
            SortParam sortParam
    ) {
        Pageable pageable = paginationParam.toPageable(sortParam.toSort());
        Page<String> sequencePage = rlooprService.findAllTags(pageable);
        return ResponsePage.responsePage(sequencePage);
    }

    @GetMapping("/{id}/average/length")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get average rloop length for single analysis")
    public Integer getAvgRloopLen(@PathVariable UUID id)
    {
        return rlooprService.avgRloopLen(id);
    }

    @GetMapping("/{id}/total/length/{model}")
    @ResponseStatus(HttpStatus.OK)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Get total rloop length for single analysis")
    public Integer getTotalRloopLen(@PathVariable UUID id, @PathVariable String model)
    {
        return rlooprService.totalRloopLen(id, model);
    }

    @PutMapping("/{id}/tags")
    @ResponseStatus(HttpStatus.ACCEPTED)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Modify tags")
    public Response<Rloop> modifyTags(
            @PathVariable UUID id,
            @RequestBody TagModifyRequest request) {
        Rloop analyse = rlooprService.modifyTags(id, request.getTags());
        return Response.response(analyse);
    }

    @DeleteMapping("/{id}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("isAuthenticated()")
    @ApiOperation("Delete one R-loop-R analysis by ID")
    public void deleteRloopRAnalysis(@PathVariable UUID id) {
        
        rlooprService.delete(id);
    }

    @GetMapping(
        path = "/{id}/rloopr.csv",
        produces = MediaType.TEXT_PLAIN_VALUE
    )
    @ResponseStatus(HttpStatus.OK)
    @ApiOperation("Get CSV file.")
    public ResponseEntity<String> getCSV(@PathVariable UUID id)
    {
        String csv = exporterService.exportData(id, Analysis.RLOOP, Output.CSV);

        HttpHeaders headers = new HttpHeaders();
        headers.setContentDisposition(ContentDisposition
                        .builder("attachment")
                        .filename(id + ".csv")
                        .build());
        headers.setContentType(MediaType.valueOf("text/csv"));
        headers.setContentLength(csv.length());

        return new ResponseEntity<>(csv, headers, HttpStatus.OK);
    }

    @GetMapping(
        path = "/{id}/rloopr.bedgraph",
        produces = MediaType.TEXT_PLAIN_VALUE
    )
    @ResponseStatus(HttpStatus.OK)
    @ApiOperation("Get BEDGRAPH file.")
    public ResponseEntity<String> getBedgraph(@PathVariable UUID id)
    {
        String bedgraph = exporterService.exportData(id, Analysis.RLOOP, Output.BEDGRAPH);

        HttpHeaders headers = new HttpHeaders();
        headers.setContentDisposition(ContentDisposition
                        .builder("attachment")
                        .filename(id + ".bedgraph")
                        .build());
        headers.setContentType(MediaType.valueOf("text/bedgraph"));
        headers.setContentLength(bedgraph.length());

        return new ResponseEntity<>(bedgraph, headers, HttpStatus.OK);
    }
}
